﻿using System;
using System.Collections.Generic;
using System.Globalization;
using System.Windows;
using System.Windows.Controls;
using System.Windows.Media;
using System.Windows.Resources;
using System.Windows.Shapes;
using System.Xml;
using Telerik.Windows.Controls;
using Telerik.Windows.Controls.Map;

namespace MapVirtualizationLayer
{
    public partial class MainWindow : Window
    {
        public MainWindow()
        {
            InitializeComponent();

            this.radMap.Provider = new OpenStreetMapProvider();
            this.virtualizationLayer.VirtualizationSource = new MapVirtualizationSource(new ExampleViewModel());
        }
    }

    public class MapVirtualizationSource : IMapVirtualizationSource
    {
        private bool isFirstRequest = true;
        private RadMap map;
        private ExampleViewModel dataContext;

        public MapVirtualizationSource(ExampleViewModel dataContext)
        {
            this.dataContext = dataContext;
        }

        public void BackgroundItemsRequest(object sender, BackgroundItemsRequestEventArgs eventArgs)
        {
            if (this.isFirstRequest)
            {
                this.map = eventArgs.Layer.MapControl;
                this.isFirstRequest = false;
            }

            double minZoom = eventArgs.MinZoom;

            // Coercing (normalizing) the map portion of the request in normal [-180; 180] longitude range.
            Location upperLeft = map.GetCoercedLocation(eventArgs.UpperLeft);
            Location lowerRight = map.GetCoercedLocation(eventArgs.LowerRight);

            if (this.dataContext == null)
                return;

            if (minZoom == 3)
            {
                // request areas.
                List<StoreLocation> list = this.dataContext.GetStores(
                    upperLeft.Latitude,
                    upperLeft.Longitude,
                    lowerRight.Latitude,
                    lowerRight.Longitude,
                    StoreType.Area);

                dataContext.SetStores(list, eventArgs);
            }

            if (minZoom == 9)
            {
                // request areas
                List<StoreLocation> list = this.dataContext.GetStores(
                    upperLeft.Latitude,
                    upperLeft.Longitude,
                    lowerRight.Latitude,
                    lowerRight.Longitude,
                    StoreType.Store);

                this.dataContext.SetStores(list, eventArgs);
            }
        }
    }

    public class StoreTemplateSelector : DataTemplateSelector
    {
        public StoreTemplateSelector()
        {
        }

        public DataTemplate AreaTemplate
        {
            get;
            set;
        }

        public DataTemplate MarketTemplate
        {
            get;
            set;
        }

        public DataTemplate StoreTemplate
        {
            get;
            set;
        }

        public override DataTemplate SelectTemplate(object item, DependencyObject container)
        {
            StoreLocation store = item as StoreLocation;
            if (store != null)
            {
                switch (store.StoreType)
                {
                    case StoreType.Area:
                        return this.AreaTemplate;

                    case StoreType.Market:
                        return this.MarketTemplate;

                    case StoreType.Store:
                    default:
                        return this.StoreTemplate;
                }
            }

            return null;
        }
    }

    public class ExampleViewModel : ViewModelBase
    {
        private MapProviderBase mapProvider;

        public ExampleViewModel()
        {
            SetProvider();

            this.ZoomLevel = 3;
            this.Center = new Location(37.684297, -99.06924);
        }

        private void SetProvider()
        {
            OpenStreetMapProvider provider = new OpenStreetMapProvider();
            provider.Layer = OpenStreetMapLayer.Humanitarian;

            this.mapProvider = provider;
        }

        public MapProviderBase Provider
        {
            get
            {
                return this.mapProvider;
            }
        }

        private int _zoomLevel;
        public int ZoomLevel
        {
            get
            {
                return this._zoomLevel;
            }
            set
            {
                if (this._zoomLevel != value)
                {
                    this._zoomLevel = value;
                    this.OnPropertyChanged("ZoomLevel");
                }
            }
        }

        private Location _center;
        public Location Center
        {
            get
            {
                return this._center;
            }
            set
            {
                if (this._center != value)
                {
                    this._center = value;
                    this.OnPropertyChanged("Center");
                }
            }
        }
        internal List<StoreLocation> GetStores(double upperLeftLat, double upperLeftLong, double lowerRightLat, double lowerRightLong, StoreType storeType)
        {
            List<StoreLocation> locations = new List<StoreLocation>();

            StreamResourceInfo streamInfo = Application.GetResourceStream(
                new Uri("/MapVirtualizationLayer;component/StoresLocation.xml", UriKind.Relative));

            XmlDocument document = new XmlDocument();
            document.Load(streamInfo.Stream);

            string latLonCondition = "[number(@Latitude) < " + Math.Round(upperLeftLat, 5).ToString(CultureInfo.InvariantCulture) +
                " and number(@Latitude) > " + Math.Round(lowerRightLat, 5).ToString(CultureInfo.InvariantCulture) +
                " and number(@Longitude) > " + Math.Round(upperLeftLong, 5).ToString(CultureInfo.InvariantCulture) +
                " and number(@Longitude) < " + Math.Round(lowerRightLong, 5).ToString(CultureInfo.InvariantCulture) + "]";

            switch (storeType)
            {
                case StoreType.Area:
                    {
                        XmlNodeList nodeList = document.SelectNodes("/StoresLocation/Area" + latLonCondition);
                        foreach (XmlNode node in nodeList)
                        {
                            XmlElement element = (XmlElement)node;

                            locations.Add(new StoreLocation(
                                Convert.ToDouble(element.GetAttribute("Latitude"), CultureInfo.InvariantCulture),
                                Convert.ToDouble(element.GetAttribute("Longitude"), CultureInfo.InvariantCulture),
                                element.GetAttribute("Name"), StoreType.Area));
                        }
                    }
                    break;

                case StoreType.Store:
                    {
                        XmlNodeList nodeList = document.SelectNodes("/StoresLocation/Area/*" + latLonCondition);
                        foreach (XmlNode node in nodeList)
                        {
                            XmlElement element = (XmlElement)node;

                            locations.Add(new StoreLocation(
                                Convert.ToDouble(element.GetAttribute("Latitude"), CultureInfo.InvariantCulture),
                                Convert.ToDouble(element.GetAttribute("Longitude"), CultureInfo.InvariantCulture),
                                element.GetAttribute("Name"),
                                element.LocalName == "Market" ? StoreType.Market : StoreType.Store));
                        }
                    }
                    break;
            }

            return locations;
        }

        /// <summary>
        /// Callback of the GetStores async call.
        /// The method uses the web service response to building objects on the dynamic layer.
        /// </summary>
		internal void SetStores(List<StoreLocation> list, BackgroundItemsRequestEventArgs request)
        {
            if (list.Count == 0)
                return;

            foreach (StoreLocation storeLocation in list)
            {
                // Shifts the store to the current portion of the request.
                storeLocation.Longitude = this.TryGetLongitudeMatchInRange(storeLocation.Longitude, request.UpperLeft.Longitude, request.LowerRight.Longitude);
            }

            request.CompleteItemsRequest(list);
        }

        private double TryGetLongitudeMatchInRange(double longitude, double left, double right)
        {
            if (left <= longitude && longitude <= right)
                return longitude;

            double range = this.mapProvider.SpatialReference.MaxLongitude - this.mapProvider.SpatialReference.MinLongitude;

            while (longitude < left)
            {
                longitude += range;
            }
            while (longitude > right)
            {
                longitude -= range;
            }

            return longitude;
        }
    }

    public class StoreLocation : ViewModelBase
    {
        private double _lattitude;
        private double _longitude;
        private string _storeName;
        private StoreType _storeType;

        /// <summary>
        /// Initializes a new instance of the StoreLocation class.
        /// </summary>
        public StoreLocation()
        {
        }

        /// <summary>
        /// Initializes a new instance of the StoreLocation class.
        /// </summary>
        /// <param name="latitude">Latitude.</param>
        /// <param name="longitude">Longitude.</param>
        /// <param name="storeName">Store name.</param>
        /// <param name="storeType">Store type.</param>
        public StoreLocation(double latitude, double longitude, string storeName, StoreType storeType)
        {
            this.Latitude = latitude;
            this.Longitude = longitude;
            this.StoreName = storeName;
            this.StoreType = storeType;
        }

        public double Latitude
        {
            get
            {
                return this._lattitude;
            }
            set
            {
                Location old = this.Location;
                this._lattitude = value;
                this.OnPropertyChanged("Latitude");
                this.OnPropertyChanged("Location");
            }
        }

        public double Longitude
        {
            get
            {
                return this._longitude;
            }
            set
            {
                Location old = this.Location;

                this._longitude = value;
                this.OnPropertyChanged("Longitude");
                this.OnPropertyChanged("Location");
            }
        }

        public Location Location
        {
            get
            {
                return new Location(this.Latitude, this.Longitude);
            }
        }

        public string StoreName
        {
            get
            {
                return this._storeName;
            }
            set
            {
                this._storeName = value;
                this.OnPropertyChanged("StoreName");
            }
        }

        public StoreType StoreType
        {
            get
            {
                return this._storeType;
            }
            set
            {
                
                this._storeType = value;
                this.OnPropertyChanged("StoreType");
            }
        }
    }

    public enum StoreType
    {
        Area,
        Market,
        Store
    }
}
